@@ -2,9 +2,10 @@ import logging
 import os
 import shutil
 from pathlib import Path
-from typing import List
+from typing import List, Dict, Optional
 
 from kodexa.model import Document
+from kodexa.pipeline import PipelineContext
 
 logger = logging.getLogger('kodexa-stores')
 
@@ -144,16 +145,21 @@ class TableDataStore:
 
     :param columns: a list of the column names (default to dynamic)
     :param rows: initial set of rows (default to empty)
+    :param source_documents: initial dictionary of document UUID to row links (default to empty)
 
     """
 
-    def __init__(self, columns=None, rows=None):
+    def __init__(self, columns=None, rows=None, source_documents=None):
+        if source_documents is None:
+            source_documents = {}
         if rows is None:
             rows = []
         if columns is None:
             columns = []
-        self.columns = columns
-        self.rows = rows
+        self.columns: List[str] = columns
+        self.rows: List[List] = rows
+        self.source_documents: Dict[str, Dict] = source_documents
+        self.pipeline_context: Optional[PipelineContext] = None
 
     """
     Return the store as a dict for serialization
@@ -173,7 +179,8 @@ class TableDataStore:
             "type": "table",
             "data": {
                 "columns": self.columns,
-                "rows": self.rows
+                "rows": self.rows,
+                "source_documents": self.source_documents
             }
         }
 
@@ -188,6 +195,9 @@ class TableDataStore:
         else:
             return pd.DataFrame(self.rows, columns=self.columns)
 
+    def set_pipeline_context(self, pipeline_context: PipelineContext):
+        self.pipeline_context = pipeline_context
+
     def add(self, row):
         """
         Writes a row to the Data Store
@@ -196,6 +206,12 @@ class TableDataStore:
         """
         self.rows.append(row)
 
+        if self.pipeline_context and self.pipeline_context.current_document:
+            current_document = self.pipeline_context.get_current_document()
+            if current_document.uuid not in self.source_documents:
+                self.source_documents[current_document.uuid] = {"metadata": current_document.metadata, "rows": []}
+            self.source_documents[current_document.uuid]["rows"].append(len(self.rows) - 1)
+
     def count(self):
         """
         Returns the number of rows in the store
@@ -225,6 +241,7 @@ class DictDataStore:
         if dicts is None:
             dicts = []
         self.dicts = dicts
+        self.pipeline_context = None
 
     """
     Return the store as a dict for serialization
@@ -246,6 +263,9 @@ class DictDataStore:
             }
         }
 
+    def set_pipeline_context(self, pipeline_context):
+        self.pipeline_context = pipeline_context
+
     def add(self, dict):
         """
         Writes a dict to the Data Store
